/****************************************************************************
 *
 *            Name:   ZV-Zip Viewer, Version 1.00
 *        Synopsis:   ZV zipfile
 *
 *     Description:   This program demonstrates an alternate way of
 *                    reading file information from a Zip File by getting
 *                    the number of files, and the offset of the central
 *                    directory, from the "end of central directory header"
 *                    and reading the file information from the central
 *                    directory, instead of reading thru all of the local
 *                    file headers before getting to the central directory.
 *                    (As done in ZIPV by S.H. Smith)
 *
 *      Written By:   Ronald C. Bieber        11-Jul-89
 *   Compiled with:   Microsoft C Version 5.1 (can be compiled with 
 *                    Turbo C 2.0, though)
 *
 *      Do not distribute modified versions without my permission.
 *      Do not remove or alter this notice or any other copyright notice.
 *      If you use this in your own program you must distribute source code.
 *      Do not use any of this in a commercial product.
 *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>

#ifdef __TURBOC__
    #include <alloc.h>
#else
    #include <malloc.h>
#endif

#define ARRAYSIZE   2048

#define ZIP_LOCAL_SIG   0x04034b50          /* local file header signature */
#define CENTRAL_SIG     0x02014b50          /* central directory structure signature */
#define END_SIG         0x06054b50          /* end of central directory signature */

#pragma pack(1)
struct Central_Dir {
    unsigned long   sig;                    /* local file header signature */
    unsigned short  made_by;                /* version made by */
    unsigned short  version;                /* version needed to extract */
    unsigned short  gen;                    /* general purpose bit flag */
    unsigned short  method;                 /* compression method */
    unsigned short  time;                   /* last mod file time */
    unsigned short  date;                   /* last mod file date */
    unsigned long   crc;                    /* CRC-32 */
    unsigned long   csize;                  /* compressed file size */
    unsigned long   size;                   /* uncompressed file size */
    unsigned short  fnamesize;              /* size of filename field */
    unsigned short  extra_field_len;        /* size of extra field */
    unsigned short  file_comment_length;    /* size of file comment */
    unsigned short  disk_no_start;          /* starting disk number */
    unsigned short  internal_attr;          /* internal file attributes */
    unsigned long   external_attr;          /* external file attributes */
    unsigned long   rel_offset_local_hdr;   /* relative offset to local file header */
};

struct End_Of_Central_Dir {
    unsigned long   sig;
    unsigned short  disk_no;
    unsigned short  disk_with_central_dir_start;
    unsigned short  no_of_entries_this_disk;
    unsigned short  tot_entries;
    unsigned long   size_of_central_dir;
    unsigned long   offset_to_start_of_central_dir;
    unsigned short  comment_length;
};
#pragma pack()

union sigtest {
    unsigned long   sig;
    char            ssig[4];
};

/*******************
 *
 *      Function:     Stowage
 *      Synopsis:     stowage(csize, size);
 *                    long    csize;      size of file after compression
 *                    long    size;       original size of file
 *
 *   Description:     This function figures the percentage of compression
 *                    performed on a given file.
 *
 ********************
int stowage(csize, size)
long    csize,
        size;
{
    float   stow = 0.00;

    if (size == csize)
        return(0);

    stow = ( ((float) size - (float) csize) / (float) size) * 100;
    
    return( (int) stow );
}

/********
 *
 *  Main Program
 *
 *******/
 
main(argc, argv)
int     argc;
char    **argv;
{
    int     fh,
            i;                              /* count of entries processed */
    struct  End_Of_Central_Dir  *eocd;
    struct  Central_Dir         *cd;
    char    fname[80],
            *buff,
            *p;
    long    flen,
            totsize = 0,
            totcsize = 0;
    union   sigtest sg;

    static char    *methods[6] = {
                        "Stored",
                        "Shrunk",
                        "Reduced-1",
                        "Reduced-2",
                        "Reduced-3",
                        "Reduced-4"
                        };

    printf("ZV-Zip Viewer, Version 1.00, (C) Copr 1989, Ronald C. Bieber\n\n");

    if (argc < 2) {
        printf("Specify ZIP file to view.\n");
        exit(1);
    }

    if ((fh = open(*(++argv), O_RDONLY | O_BINARY)) == -1) {
        perror(*argv);
        exit(1);
    }

    if ((buff = (char *) malloc(ARRAYSIZE)) == 0) {
        fprintf(stderr, "Out of memory ...\n");
        exit(1);
    }

    if ((eocd = (struct End_Of_Central_Dir *) malloc(sizeof(struct End_Of_Central_Dir))) == NULL) {
        fprintf(stderr, "Out of memory ...\n");
        free(buff);
        exit(1);
    }

    if ((cd = (struct Central_Dir *) malloc(sizeof(struct Central_Dir))) == NULL) {
        fprintf(stderr, "Out of memory ...\n");
        free(buff);
        free(eocd);
        exit(1);
    }

    lseek(fh, filelength(fh) - ARRAYSIZE, SEEK_SET);

    i = 0;

    p = buff + read(fh, buff, ARRAYSIZE);

    while (p > buff) {
    strncpy(sg.ssig, p, 4);
        if (sg.sig == END_SIG)
            break;
        else {
            p--;
            i++;
        }
    }

    if (p <= buff) {
        fprintf(stderr, "%s is not a zipfile or is corrupted.\n", *argv);
        exit(1);
    }

    free(buff);

    flen = filelength(fh);
    flen -= i;

    lseek(fh, flen, SEEK_SET);
    
    read(fh, eocd, sizeof(struct End_Of_Central_Dir));
    lseek(fh, (long) eocd->offset_to_start_of_central_dir, SEEK_SET);

    printf("Listing of ZipFile:     %s\n\n", strupr(*argv));

    printf("Length   Method      Size   Ratio   Date     Time   Name\n");
    printf("-------  ------      ------ -----   ----     ----   ----\n");

    for (i = 0; i < eocd->tot_entries; i++) {
    
        if (read(fh, cd, (long) sizeof(struct Central_Dir)) < sizeof(struct Central_Dir) || cd->sig != CENTRAL_SIG)
            break;
            
        read(fh, fname, cd->fnamesize);
        fname[cd->fnamesize] = 0;

        printf("%7lu  %-9.9s  %7lu ", cd->size, methods[cd->method], cd->csize);
        printf("(%2d%%) ", stowage(cd->csize, cd->size));

        totsize += cd->size;
        totcsize += cd->csize;

        printf("%2.2d-%2.2d-%2d ",  (cd->date >> 5) & 0x0f, cd->date & 0x1f,
                                    ((cd->date >> 9) & 0x7f) + 80);

        printf("%2.2d:%2.2d:%2.2d ", (cd->time >> 11) & 0x1f, (cd->time >> 5) & 0x3f,
                                      (cd->time & 0x1f) * 2);

        printf("%-13.13s\n", strupr(fname));
        lseek(fh, (long) (cd->extra_field_len + cd->file_comment_length), SEEK_CUR);
    }

    printf("--------           -------- -----                   ----\n");
    printf("%8lu           %8lu (%d%%)                   %4d\n", totsize,
                                                                 totcsize,
                                                                 stowage(totcsize, totsize),
                                                                 eocd->tot_entries);

    free(eocd);
    free(cd);
    close(fh);
}
